#!/bin/env python
"""
Configuration file parser.

See the README.txt file (part of this source code distribution) for usage,
contact info, how to cite, etc.

Copyright 2010 Andrew V. Uzilov and The Regents of the University of California
All Rights Reserved

This file is part of FragSeq_v0.0.1.

FragSeq_v0.0.1 is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FragSeq_v0.0.1 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FragSeq_v0.0.1.  If not, see <http://www.gnu.org/licenses/>.
"""

import re
import sys


#===============================================================================
# Classes.
#===============================================================================

class ConfigParseError (EnvironmentError):
    """Indicates a parse error of the config file that we are loading."""


class Config():
    """Represents a configuration loaded from a file.
    """

    # Regexps for parsing config files.
    reConfigLine  = re.compile (r'^\s*(\S+)\s+(\S.*)$')
    reVarDefLine  = re.compile (r'^\s*%define\s+(\S+)\s+(\S.*)$')
    reCommentLine = re.compile (r'^\s*#')
    reBlankLine   = re.compile (r'^\s*$')


    class ConfigAttribute():
        """A config file attribute.
        """


    def __init__ (self, validAttrs, configFileNameOrList=None,
                  multValues='forbid', mergeConfigFiles=False):
        if not isinstance (validAttrs, dict):
            raise TypeError ("Argument 'validAttrs' to 'Config' object constructor must be of a dictionary type.")

        # Pre-initialize all valid attributes to None, so there are
        # no AttributeError issues with this Config object down the line.
        def initAttr (obj, attrDict):
            for attrName in attrDict.keys():
                if isinstance (attrDict[attrName], dict):
                    # Recurse...
                    setattr (obj, attrName, Config.ConfigAttribute())
                    initAttr (getattr (obj, attrName), attrDict[attrName])
                else:
                    # Hit the end...
                    setattr (obj, attrName, None)
        initAttr (self, validAttrs)

        # Certainly no one would be insane enough to set this...
        assert not hasattr (self, 'validAttrs')
        self.validAttrs = validAttrs

        if configFileNameOrList is not None:
            if isinstance (configFileNameOrList, list):
                # Load configs from multiple files.
                if mergeConfigFiles:
                    configLines = []
                    for fileName in configFileNameOrList:
                        configLines += [l for l in open (fileName, 'r')]
                    try:
                        self.addConfigFromLines (
                            configLines, multValues)
                    except ConfigParseError, e:
                        raise ConfigParseError (
                            'Error loading the concatenation of these config file: %s\n%s'
                            % (configFileNameOrList, str (e)))
                else:
                    for fileName in configFileNameOrList:
                        self.addConfigFromFile (fileName, multValues)
            else:
                # Load config from a single file.
                self.addConfigFromFile (
                    configFileNameOrList, multValues)


    def __str__ (self):
        def getAttr (obj, attrDict):
            retLocal = []
            for attrName in attrDict.keys():
                if isinstance (attrDict[attrName], dict):
                    # Recurse.
                    subList = getAttr (
                        getattr (obj, attrName), attrDict[attrName])
                    retLocal += [attrName + ':' + s for s in subList]
                else:
                    # Hit the end.
                    retLocal += [
                        '%s\t%s' %(attrName,  getattr (obj, attrName))]
            return retLocal
        ret = getAttr (self, self.validAttrs)
        ret.sort()
        return '\n'.join (ret)


    def addConfigFromFile (self, configFileName, multValues='forbid'):
        try:
            self.addConfigFromLines ([l for l in open (configFileName, 'r')],
                                     multValues)
        except ConfigParseError, e:
            raise ConfigParseError (
                'Error loading this config file: %s\n%s'
                % (configFileName, str (e)))

    
    def addConfigFromLines (self, configLines, multValues='forbid'):
        """Load tags and values from a set of lines from a config files,
        appending them to Config object as attributes.

        'configLines' must be an iterable, where each value returned is a
        string.

        ConfigParseError with details will be raised on failure.

        If we come across a setting for a tag that was already set, a
        ConfigParseError is raised.  This can be turned off by setting the
        allowValueOverride option to True, allowing override of previous tag
        values.
        """
        # First pass:
        # parse config file for %define variable definitions.
        # 'fileVars' maps variable names to values for this config file.
        fileVars = {}
        for line in configLines:
            matchVarDefLine = self.reVarDefLine.search (line)
            if matchVarDefLine:
                varName = matchVarDefLine.group (1)
                varVal  = matchVarDefLine.group (2).rstrip()
                #sys.stderr.write ('Saving var def (%s) = (%s)\n' % (varName, varVal))  #d!!!
                fileVars[varName] = varVal

        # Second pass:
        # expand variables and parse config file for everything else.
        lineNum = 0
        for line in configLines:
            lineNum += 1
            
            # Skip comment lines, blank lines,
            # variable definition lines (already parsed earlier).
            if self.reCommentLine.search (line) \
                    or self.reBlankLine.search (line) \
                    or self.reVarDefLine.search (line):
                continue
            
            # Expand variable names to values.
            matchConfigLine = self.reConfigLine.search (line)
            if matchConfigLine:
                tag   = matchConfigLine.group (1)
                value = matchConfigLine.group (2).rstrip()
                for varName, varVal in fileVars.iteritems():
                    #sys.stderr.write ('Trying expansion of %s to (%s)\n' % ('${%s}' % varName, varVal))  #d!!!
                    tag   = tag.replace   ('${%s}' % varName, varVal)
                    value = value.replace ('${%s}' % varName, varVal)
            else:
                raise ConfigParseError ('Malformed line %d.' % lineNum)
            
            # Safety.
            # Loop through fields in the tag while also recursing through
            # dictionary/tree of allowed tags/fields, to make sure the tag
            # structure is valid.
            tagFields = tag.split (':')
            vaLevel   = self.validAttrs
            for field in tagFields:
                if isinstance (vaLevel, dict):
                    if not vaLevel.has_key (field):
                        raise ConfigParseError (
                            'Invalid tag field "%s" on line %d.  Valid tag fields at this position are: %s.' \
                            % (field, lineNum, ', '.join (
                                ['"%s"' % f for f in vaLevel.keys()]
                                )))
                    else:
                        vaLevel = vaLevel[field]  # Recurse.
                else:
                    # 'vaLevel' is not a dictionary.
                    raise ConfigParseError (
                        'Invalid tag field "%s" on line %d.  There should be no tag fields at this position.' \
                            % (field, lineNum))
            
            # Infer how to interpret and safety-check 'value' based on its
            # tag's entry (to which 'vaLevel' now points) in 'validAttrs'.
            if vaLevel is not None:
                try:
                    # Let's assume 'vaLevel' gives us an iterable of allowed
                    # values for this tag.  Check to make sure value specified
                    # by config file is a valid setting.
                    if value not in vaLevel:
                        raise ConfigParseError (
                            'Invalid value "%s" for tag "%s" on line %d.  The value here must be one of: "%s".' \
                            % (value, tag, lineNum, vaLevel))
                except TypeError:
                    # Oops, 'vaLevel' is not an iterable.
                    # Must instead be specifying a Python type.  Try to
                    # convert value in config file to this type.
                    try:
                        value = vaLevel (value)
                    except ValueError:
                        raise ConfigParseError (
                            'Invalid value "%s" for tag "%s" on line %d.  The value here must be of type "%s".' \
                            % (value, tag, lineNum, vaLevel))
            
            # Save the value of the setting.
            # For example, the value from config file line:
            #     infile:reads:ss:seed    fileName
            # will be saved like
            #     self.infile.reads.ss.seed = fileName

            # Find the ConfigAttribute object we need (save to 'lastAttr').
            # 'value' will be saved into one of its attributes.
            lastAttr = self
            for field in tagFields[:-1]:
                if not hasattr (lastAttr, field):
                    setattr (lastAttr, field, Config.ConfigAttribute())
                    assert False, "Cyant happen."
                lastAttr = getattr (lastAttr, field)
            
            # Save the value.
            if getattr (lastAttr, tagFields[-1]) is not None:
                # Value has already been set; what to do?
                if multValues == 'forbid':
                    # Overwriting previous values not allowed.
                    raise ConfigParseError (
                        'Trying to set a duplicate value for tag "%s" on line %d.  Value for this tag was already specified earlier as "%s".' \
                        % (tag, lineNum, getattr (lastAttr, tagFields[-1])))
                elif multValues == 'append':
                    # Append to previous value(s).
                    if isinstance (getattr (lastAttr, tagFields[-1]), list):
                        # Already a list.
                        getattr (lastAttr, tagFields[-1]).append (value)
                    else:
                        # Make into a list.
                        setattr (lastAttr, tagFields[-1], [
                            getattr (lastAttr, tagFields[-1]), value])
                elif multValues == 'overwrite':
                    # Overwrite previous value.
                    setattr (lastAttr, tagFields[-1], value)
                else:
                    assert False, "Cyant happen!"
            else:
                setattr (lastAttr, tagFields[-1], value)

